/**
 * ToonLoop :: Realtime stop motion performance tool. 
 * @author Alexandre Quessy <alexandre@quessy.net>
 * 
 * Usage : 
 *  - Press space bar to grab a frame.
 *  - Press DELETE or BACKSPACE to delete the last frame.
 *  - Press 'r' to reset and start an empty movie.
 *  - Press 's' to save as a QuickTime movie.
 *  - Press 'p' to open the Quicktime video camera settings dialog. (if available)
 *  - Press 'a' to toggle on/off the auto recording.
 *  - Press 'q' to quit
 * 
 * Installation :
 *  - Run the application in the application.macosx/ or application.windows/ folder.
 *    It should work. If not, you need to compile the sketch as follow:
 *  - You need Processing (http://processing.org). 
 *  - Put the toonloop folder in ~/Documents/Processing/
 *    (where "~" represents your home directory, such as "/Users/alexandre/")
 *  - Change Processing::Preferences::Set maximum available memory to...
 *    to a higher value, such as 500 Mb at least.
 *  - Set LOOP_MAX_NUM_FRAME to the desired number of frames.
 *  - Run the sketch.
 *  - Quit the applet to stop.
 *  - Close the window and start over to create a new movie.
 *  - Look in ~/Documents/Processing/toonloop_07/ for the saved movie files.
 * 
 * @version 0.7 Improving usability
 *              Added features: delete frames, reset movie, auto record.
 *              Configuration with variables.
 *              Frames numbering starts at 1. 
 * @date 2008-09-27
 * @license GNU General Public License version 2
 *          Contact the author if you want a commercial version and license of this tool.
 * @url http://alexandre.quessy.net/?q=toonloop
 */
import processing.video.*;

// ---------------- CONFIGURATION -------------------//
int LOOP_MAX_NUM_FRAME = 500; // Max number of frames.
int LOOP_WIDTH = 320;
int LOOP_HEIGHT = 240;
int FRAME_RATE = 12;
int TEXT_FONT_SIZE = 24;

/* Choose only one of the followings: */
//int LOOP_CODEC = MovieMaker.H263;
//int LOOP_CODEC = MovieMaker.H264;
//int LOOP_CODEC = MovieMaker.MOTION_JPEG_A;
int LOOP_CODEC = MovieMaker.MOTION_JPEG_B;
//int LOOP_CODEC = MovieMaker.ANIMATION;

/* Choose only one of the followings: */
//int LOOP_QUALITY = MovieMaker.HIGH;
int LOOP_QUALITY = MovieMaker.BEST;
//int LOOP_QUALITY = MovieMaker.LOSSLESS;
// -------------------------------------------------//

int captureFrameNum = 0; //the next captured frame number ... might wrap around
int playFrameNum = 0;
int isFlashing = 0;
int is_auto_recording = 0;
int is_displaying_saved_message = 0; // also a counter
int SAVED_MESSAGE_DURATION = 30; // frames
String saved_file_name;
float SAVED_MESS_SIZE_RATIO = 0.5;

PImage[] images = new PImage[LOOP_MAX_NUM_FRAME];
Capture cam;
MovieMaker movieOut; 
PFont font;

void setup() 
{
  size(LOOP_WIDTH*2, LOOP_HEIGHT*2);
  frameRate(FRAME_RATE);

  println("Available cameras :");
  print(Capture.list()); 
  println("Please wait while it is loading...");
  println("");
  println("Welcome to ToonLoop ! The realtime stop motion performance tool.");
  println(")c( Alexandre Quessy 2008");
  println("http://alexandre.quessy.net");
  
  for (int i = 0; i < LOOP_MAX_NUM_FRAME; i++) 
  {
    //images[i] = new PImage(LOOP_WIDTH, LOOP_HEIGHT);
  }
  cam = new Capture(this, LOOP_WIDTH, LOOP_HEIGHT);
  
  font = loadFont("CourierNewPSMT-24.vlw");
  
  println("Done loading.");
}

void draw() 
{
  background(0);
  
  if (cam.available() == true) 
  {
    cam.read();
  }
  
  // Live camera image on the left:
  // with message "AUTO RECORDING" if recording
  noTint(); // alpha 100%
  textFont(font, TEXT_FONT_SIZE); // size in pixels
  if (is_auto_recording == 1) {
    addFrame();
    fill(255,0,0,255);
    //String warningText = "Auto recording on. Press 'a' to stop.";
    String warningText = "AUTO RECORDING";
    text(warningText, (int)(LOOP_WIDTH/4),(int)(LOOP_HEIGHT*1.8)); 
  } else {
    fill(255,255,255,255);
    if (isFlashing == 1) 
    {
      isFlashing = 0;
      rect(0,LOOP_HEIGHT/2,LOOP_WIDTH,LOOP_HEIGHT);
      tint(255, 191); // alpha 75%
    } 
  }
  image(cam,0,LOOP_HEIGHT/2);
  
  // Image at the current offset in the loop on the right:
  noTint(); // alpha 100%
  if (captureFrameNum > 0) 
  { 
    image(images[playFrameNum], LOOP_WIDTH, (int)(LOOP_HEIGHT/2));
    loopFrame();
  }
  
  // Frame number under the images :
  fill(127);
  textFont(font, TEXT_FONT_SIZE); // size in pixels
  if (captureFrameNum > 0)
  {
    text(""+captureFrameNum, (int)(LOOP_WIDTH/2),(int)(LOOP_HEIGHT*1.6));
    text(""+(playFrameNum+1), (int)(LOOP_WIDTH*3/2),(int)(LOOP_HEIGHT*1.6)); 
  }
  
  // SAVED message
  textFont(font, (int)(TEXT_FONT_SIZE* SAVED_MESS_SIZE_RATIO)); // size in pixels
  if (is_displaying_saved_message > 0)
  {
    is_displaying_saved_message--; // decrement duration
    int x = (int)(LOOP_WIDTH/4);
    int y = (int)(LOOP_HEIGHT*1.8);
    fill(255,0,0,255);
    text("Saved to "+saved_file_name + ". ("+LOOP_WIDTH +"x"+ LOOP_HEIGHT+")", x,y);
  }
}

// controls
void keyPressed() {
  switch (key)
  {
    case ' ': // ADD FRAME
      if (is_auto_recording == 0) {
        addFrame();
      }
      break;
    case 'p': // SETTINGS
      cameraPreferences();
      break;
    case 'r': // RESET
      resetMovie();
      break;
    case 's':
      if (is_auto_recording == 0) {
        saveMovie();
      }
      break;
    case BACKSPACE: // DELETE ONE FRAME
    case DELETE:
      deleteFrame();
      break;
    case 'a': // AUTO RECORD
      toggleAutoRecording();
      break;
    case ESC: // FULL SCREEN
      // todo (OpenGL)
      break;
    case 'q': // QUIT APPLICATION
      exit();
      break;
  }
}

void toggleAutoRecording()
{
  if (is_auto_recording ==0) {
    is_auto_recording = 1;
  } else {
    is_auto_recording = 0;
  }
}

void loopFrame() 
{
  if (playFrameNum < captureFrameNum - 1) 
  {
    playFrameNum++;
  } else {
    playFrameNum = 0;
  }
}

void addFrame() 
{
  if (captureFrameNum < LOOP_MAX_NUM_FRAME) 
  {
    // We use new here, because it was not possible to overwrite an existing image.
    images[captureFrameNum] = new PImage(LOOP_WIDTH, LOOP_HEIGHT);
    images[captureFrameNum].copy(cam,0,0,LOOP_WIDTH,LOOP_HEIGHT,0,0,LOOP_WIDTH,LOOP_HEIGHT);
    
    captureFrameNum++;
    isFlashing = 1;
  } 
  else {
    println("Reached max number of frames : " + LOOP_MAX_NUM_FRAME);
  }
}
void deleteFrame()
{
  // the fomer frame goes to the garbage collector
  if (captureFrameNum > 0) 
  {
    captureFrameNum--;
  }
}
void cameraPreferences()
{
  println("Opening the settings window. (if available)");
  cam.settings();
}
void resetMovie() 
{
  captureFrameNum = 0;
}
void saveMovie() 
{
  // Create MovieMaker object with size, filename,
  // compression codec and quality, framerate
  String file_name;
  if (captureFrameNum == 0) {
    println("no frame to save!");
  } 
  else {
    file_name = "toonloop_"+String.valueOf(year())
      +"_"+String.valueOf(month())
        +"_"+String.valueOf(day())
          +"_"+String.valueOf(hour())
            +"h"+String.valueOf(minute())
              +"m"+String.valueOf(second())
                +".mov";
    println("saving to "+file_name+" :");
    
    // enable saved message.
    is_displaying_saved_message = SAVED_MESSAGE_DURATION;
    saved_file_name = file_name;
    
    movieOut = new MovieMaker(this, LOOP_WIDTH, LOOP_HEIGHT, 
    file_name, FRAME_RATE, LOOP_CODEC, LOOP_QUALITY);
    for (int i = 0; i < captureFrameNum; i++) 
    {
      print(i+" ");
      movieOut.addFrame(images[i].pixels, LOOP_WIDTH, LOOP_HEIGHT);
    }
    println();
    movieOut.finish();
    println("done saving");
    movieOut = null;
  }
}